@@ -5,7 +5,7 @@ cd examples
 for i in *.v
 do
   coqdoc -s --no-lib-name -parse-comments --no-index --utf8 --interpolate --html \
-      --external http://github.com/mattam82/Coq-Equations/tree/master Equations \
+      --external http://github.com/mattam82/Coq-Equations/tree/main Equations \
       -Q ../theories Equations -R . Examples -d ../../equations-www/examples $i
 done
 
