@@ -30,7 +30,7 @@ type int_data = {
   flags : flags;
   program_mode : bool;
   intenv : Constrintern.internalization_env;
-  notations : Vernacexpr.decl_notation list
+  notations : Vernacexpr.notation_declaration list
 }
 
 exception Conflict
@@ -56,7 +56,7 @@ let decompose_rel evd c =
   | _ -> None
 
 let rec unify env evd flex g x y =
-  if check_conv env evd x y then id_subst g
+  if is_conv_leq env evd x y then id_subst g
   else
     match decompose_rel evd x with
     | Some i -> 
@@ -72,8 +72,8 @@ let rec unify env evd flex g x y =
           single_subst env evd i (PInac x) g
         else raise Stuck
       | None ->
-        let (c, l) = decompose_app evd x 
-        and (c', l') = decompose_app evd y in
+        let (c, l) = decompose_app_list evd x
+        and (c', l') = decompose_app_list evd y in
         if isConstruct evd c && isConstruct evd c' then
           if eq_constr evd c c' then
             unify_constrs env evd flex g l l'
@@ -310,7 +310,7 @@ let is_wf_ref id rec_type =
     | Some (Logical (nargs, (_, id'))) -> 
       if Id.equal id id' then Some nargs else None 
     | _ -> None
-  in CList.find_map aux rec_type
+  in CList.find_map_exn aux rec_type
 
 let add_wfrec_implicits rec_type c =
   let open Glob_term in
@@ -331,7 +331,7 @@ let add_wfrec_implicits rec_type c =
                                             qm_name = Anonymous;
                                             qm_record_field = None }
                     in
-                    let newarg = GHole (Evar_kinds.QuestionMark kind, Namegen.IntroAnonymous, None) in
+                    let newarg = GHole (Evar_kinds.QuestionMark kind, Namegen.IntroAnonymous) in
                     let newarg = DAst.make ?loc newarg in
                     let before, after = List.chop nargs (mapargs args) in
                     let args' = List.append before (newarg :: after) in
@@ -449,18 +449,18 @@ let unify_type env evars before id ty after =
       Array.mapi (fun i ty ->
           let ty = of_constr ty in
           let ty = prod_applist !evars ty params in
-          let ctx, ty = decompose_prod_assum !evars ty in
+          let ctx, ty = decompose_prod_decls !evars ty in
           let ctx = 
             fold_right (fun decl acc ->
-                let (n, b, t) = to_tuple decl in
-                match n.binder_name with
-                | Name id -> let id' = next_ident_away id in
-                  (make_def (nameR id') b t :: acc)
+                let open Context.Rel.Declaration in
+                let id = match get_name decl with
+                | Name id -> next_ident_away id
                 | Anonymous ->
                   let x = Namegen.id_of_name_using_hdchar
-                      (push_rel_context acc envb) !evars t Anonymous in
-                  let id = next_ident_away x in
-                  (make_def (nameR id) b t :: acc))
+                      (push_rel_context acc envb) !evars (get_type decl) Anonymous in
+                  next_ident_away x
+                in
+                (set_name (Name id) decl :: acc))
               ctx []
           in
           let env' = push_rel_context ctx env in
@@ -639,7 +639,7 @@ let instance_of_pats env evars (ctx : rel_context) (pats : (int * bool) list) =
   let pats' =
     List.map_i (fun i id ->
         let i', _ = lookup_named_i id ctx' in
-        CList.find_map (fun (i'', hide) ->
+        CList.find_map_exn (fun (i'', hide) ->
             if i'' == i then Some (if hide then PHide i' else PRel i')
             else None) pats)
       1 subst
@@ -648,7 +648,7 @@ let instance_of_pats env evars (ctx : rel_context) (pats : (int * bool) list) =
     List.map_i (fun i decl ->
         let (id, b, t) = to_named_tuple decl in
         let i', _ = lookup_named_i id.binder_name nctx in
-        CList.find_map (fun (i'', hide) ->
+        CList.find_map_exn (fun (i'', hide) ->
             if i'' == i' then Some (if hide then PHide i else PRel i)
             else None) pats)
       1 ctx'
@@ -663,9 +663,9 @@ let push_rel_context_eos ctx env evars =
     in push_rel_context ctx env'
   else push_rel_context ctx env
 
-let split_at_eos sigma ctx =
+let split_at_eos env sigma ctx =
   List.split_when (fun decl ->
-      is_lglobal sigma coq_end_of_section (get_named_type decl)) ctx
+      is_lglobal env sigma coq_end_of_section (get_named_type decl)) ctx
 
 let pr_problem p env sigma (delta, patcs, _) =
   let env' = push_rel_context delta env in
@@ -721,7 +721,9 @@ let print_program_info env sigma programs =
 let make_fix_proto env sigma ty =
   let relevance = Retyping.relevance_of_type env sigma ty in
   let _, fixproto = get_fresh sigma coq_fix_proto in
-    relevance, mkLetIn (anonR, fixproto, Retyping.get_type_of env sigma fixproto, lift 1 ty)
+  let r = Retyping.relevance_of_term env sigma fixproto in
+  let na = make_annot Anonymous r in
+  relevance, mkLetIn (na, fixproto, Retyping.get_type_of env sigma fixproto, lift 1 ty)
 
 let compute_fixdecls_data env evd ?data programs =
   let protos = List.map (fun p ->
@@ -743,7 +745,7 @@ let interp_arity env evd ~poly ~is_rec ~with_evars notations (((loc,i),udecl,rec
   let (arity, impls') =
     let ty = match t with
       | Some ty -> ty
-      | None -> CAst.make ?loc (Constrexpr.CHole (None, Namegen.IntroAnonymous, None))
+      | None -> CAst.make ?loc (Constrexpr.CHole (None, Namegen.IntroAnonymous))
     in
     Equations_common.evd_comb1 (interp_type_evars_impls env' ?impls:None) evd ty
   in
@@ -782,7 +784,7 @@ let interp_arity env evd ~poly ~is_rec ~with_evars notations (((loc,i),udecl,rec
   let program_sort =
     let u = Retyping.get_sort_of env !evd program_orig_type in
     let sigma, sortl, sortu = nonalgebraic_universe_level_of_universe env !evd u in
-    evd := sigma; sortu
+    evd := sigma; ESorts.kind sigma sortu
   in
   let program_implicits = Impargs.compute_implicits_with_manual env !evd program_orig_type false impls in
   let () = evd := Evd.minimize_universes !evd in
@@ -900,10 +902,10 @@ let wf_fix_constr env evars sign arity sort carrier cterm crel =
   in
   let fix = norm env fix in
   let functional_type, full_functional_type =
-    let ctx, rest = Reductionops.splay_prod_n env !evars (Context.Rel.nhyps sign) functional_type in
+    let ctx, rest = Reductionops.hnf_decompose_prod_n_decls env !evars (Context.Rel.nhyps sign) functional_type in
     match kind !evars (whd_all (push_rel_context ctx env) !evars rest) with
     | Prod (na, b, concl) ->
-      let ctx', rest = Reductionops.splay_prod_assum (push_rel_context ctx env) !evars b in
+      let ctx', rest = Reductionops.hnf_decompose_prod_decls (push_rel_context ctx env) !evars b in
       let infos = Cbv.create_cbv_infos reds (push_rel_context ctx env) !evars in
       let norm = Cbv.cbv_norm infos in
       let fn_type = it_mkProd_or_LetIn rest ctx' in
@@ -923,11 +925,12 @@ let wf_fix_constr env evars sign arity sort carrier cterm crel =
 let wf_fix env evars subst sign arity sort term rel =
   let envsign = push_rel_context sign env in
   let sigma, cterm = interp_constr_evars envsign !evars term in
-  let carrier =
+  let na, carrier =
+    let r = Retyping.relevance_of_term envsign sigma cterm in
     let ty = Retyping.get_type_of envsign sigma cterm in
     let ty = nf_all envsign sigma ty in
     if noccur_between sigma 1 (length sign) ty then
-      lift (- length sign) ty
+      make_annot Anonymous r, lift (- length sign) ty
     else
       user_err_loc (Constrexpr_ops.constr_loc term,
                     str"The carrier type of the recursion order cannot depend on the arguments")
@@ -937,7 +940,7 @@ let wf_fix env evars subst sign arity sort term rel =
   let sigma, rsort = Evd.fresh_sort_in_family sigma (Lazy.force Equations_common.logic_sort) in
   let sigma, crel =
     let relty =
-      (mkProd (anonR, carrier, mkProd (anonR, lift 1 carrier, mkSort rsort)))
+      (mkProd (na, carrier, mkProd (na, lift 1 carrier, mkSort rsort)))
     in
     match rel with
     | Some rel -> interp_casted_constr_evars env sigma rel relty
@@ -1274,7 +1277,7 @@ and interp_clause env evars p data prev clauses' path (ctx,pats,ctx' as prob)
       let userc, usercty = interp_constr_in_rhs env ctx evars data None s lets (GlobConstr user) in
       match t with
       | PInac t ->
-        begin match Evarconv.unify env' !evars Reduction.CONV userc t with
+        begin match Evarconv.unify env' !evars Conversion.CONV userc t with
           | evars' -> evars := evars'
           | exception Pretype_errors.PretypeError (env, sigma, e) ->
             DAst.with_loc_val (fun ?loc _ ->
@@ -1509,7 +1512,7 @@ and interp_clause env evars p data prev clauses' path (ctx,pats,ctx' as prob)
          str "And clauses: " ++ pr_preclauses env !evars cls')
     | Some (clauses, s) ->
       let () = check_unused_clauses env !evars clauses in
-      let term, _ = term_of_tree env evars p.program_sort s in
+      let term, _ = term_of_tree env evars (ESorts.make p.program_sort) s in
       let info =
         { refined_obj = (idref, cconstr, cty);
           refined_rettyp = ty;
@@ -1531,7 +1534,7 @@ and interp_clause env evars p data prev clauses' path (ctx,pats,ctx' as prob)
 
 and interp_wheres env0 ctx evars path data s lets
     (ctx, envctx, liftn, subst)
-    (w : (pre_prototype * pre_equation list) list * Vernacexpr.decl_notation list) =
+    (w : (pre_prototype * pre_equation list) list * Vernacexpr.notation_declaration list) =
   let notations = snd w in
   let aux (data,lets,nlets,coverings,env)
       (((loc,id),udecl,nested,b,t,reca),clauses as eqs) =
