@@ -40,55 +40,64 @@ let equations_transparent = ref false
 let equations_derive_equations = ref true
 let equations_derive_eliminator = ref true
 
-let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = true;
+let depr_with_k = Deprecation.make ~since:"equations v1.2" ()
+
+let () = Goptions.declare_bool_option {
+  Goptions.optdepr  = Some depr_with_k;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Equations"; "WithK"];
   Goptions.optread  = (fun () -> false);
   Goptions.optwrite = (fun b ->
       if b then
         CErrors.user_err (str"DEPRECATED. Use flag [Equations With UIP] and introduce \
-                              an axiomn [forall A, Equations.Classes.UIP A] \
+                              an axiom [forall A, Equations.Classes.UIP A] \
                               as a type class instance using [Existing Instance] instead.")
       else simplify_withUIP := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = true;
+  Goptions.optdepr  = Some depr_with_k;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Equations"; "WithKDec"];
   Goptions.optread  = (fun () -> !simplify_withUIP);
   Goptions.optwrite = (fun b -> simplify_withUIP := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Equations"; "With"; "UIP"];
   Goptions.optread  = (fun () -> !simplify_withUIP);
   Goptions.optwrite = (fun b -> simplify_withUIP := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Equations"; "Transparent"];
   Goptions.optread  = (fun () -> !equations_transparent);
   Goptions.optwrite = (fun b -> equations_transparent := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Equations"; "With"; "Funext"];
   Goptions.optread  = (fun () -> !equations_with_funext);
   Goptions.optwrite = (fun b -> equations_with_funext := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Equations"; "Derive"; "Equations"];
   Goptions.optread  = (fun () -> !equations_derive_equations);
   Goptions.optwrite = (fun b -> equations_derive_equations := b)
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Equations"; "Derive"; "Eliminator"];
   Goptions.optread  = (fun () -> !equations_derive_eliminator);
   Goptions.optwrite = (fun b -> equations_derive_eliminator := b)
@@ -99,7 +108,8 @@ let _ = Goptions.declare_bool_option {
 let debug = ref false
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Equations"; "Debug"];
   Goptions.optread  = (fun () -> !debug);
   Goptions.optwrite = (fun b -> debug := b)
@@ -210,6 +220,19 @@ let e_type_of env evd t =
   let evm, t = Typing.type_of ~refresh:false env !evd t in
   evd := evm; t
 
+let collapse_term_qualities uctx c =
+  let nf_evar _ = None in
+  let nf_univ u = u in
+  let nf_sort s = s in
+  let nf_relevance r = match r with
+  | Sorts.Relevant | Sorts.Irrelevant -> r
+  | Sorts.RelevanceVar q ->
+    match UState.nf_qvar uctx q with
+    | QVar _ (* Hack *) | QType | QProp -> Sorts.Relevant
+    | QSProp -> Sorts.Irrelevant
+  in
+  UnivSubst.nf_evars_and_universes_opt_subst nf_evar nf_univ nf_sort nf_relevance c
+
 let make_definition ?opaque ?(poly=false) evm ?types b =
   let env = Global.env () in
   let evm = match types with
@@ -220,7 +243,9 @@ let make_definition ?opaque ?(poly=false) evm ?types b =
   in
   let evm = Evd.minimize_universes evm in
   let evm0 = evm in
-  let body = EConstr.to_constr evm b and typ = Option.map (EConstr.to_constr evm) types in
+  let to_constr c = collapse_term_qualities (Evd.evar_universe_context evm) (EConstr.to_constr evm c) in
+  let body = to_constr b in
+  let typ = Option.map to_constr types in
   let used = Vars.universes_of_constr body in
   let used' = Option.cata Vars.universes_of_constr Univ.Level.Set.empty typ in
   let used = Univ.Level.Set.union used used' in
@@ -248,7 +273,7 @@ let declare_instance id ~poly evm ctx cl args =
   let term = it_mkLambda_or_LetIn (Option.get c) ctx in
   let typ = EConstr.it_mkProd_or_LetIn t ctx in
   let cst, ecst = declare_constant id term (Some typ) ~poly evm ~kind:Decls.(IsDefinition Instance) in
-  let () = Classes.Internal.add_instance (fst cl) Hints.empty_hint_info true (GlobRef.ConstRef cst) in
+  let () = Classes.declare_instance (Global.env ()) evm (Some Hints.empty_hint_info) Hints.SuperGlobal (GlobRef.ConstRef cst) in
   cst, ecst
 
 let coq_zero = (find_global "nat.zero")
@@ -333,7 +358,7 @@ let get_fresh sigma r = new_global sigma (Lazy.force r)
 
 let get_efresh r evd = e_new_global evd (Lazy.force r)
 
-let is_lglobal sigma gr c = EConstr.isRefX sigma (Lazy.force gr) c
+let is_lglobal env sigma gr c = EConstr.isRefX env sigma (Lazy.force gr) c
 
 open EConstr
 
@@ -539,12 +564,14 @@ let unfold_head env sigma db t =
 
 let autounfold_heads db db' cl =
   Proofview.Goal.enter begin fun gl ->
+  let env = Proofview.Goal.env gl in
+  let sigma = Proofview.Goal.sigma gl in
   let eq = 
     (match cl with Some (id, _) -> pf_get_hyp_typ id gl | None -> pf_concl gl) 
   in
   let did, c' = 
     match kind (project gl) eq with
-    | App (f, [| ty ; x ; y |]) when EConstr.isRefX (project gl) (Lazy.force logic_eq_type) f ->
+    | App (f, [| ty ; x ; y |]) when EConstr.isRefX env sigma (Lazy.force logic_eq_type) f ->
       let did, x' = unfold_head (pf_env gl) (project gl) db x in
       let did', y' = unfold_head (pf_env gl) (project gl) db' y in
       did && did', EConstr.mkApp (f, [| ty ; x' ; y' |])
@@ -779,10 +806,10 @@ let decompose_indapp sigma f args =
   | _ -> f, args
 
 let e_conv env evdref t t' =
-  match Reductionops.infer_conv env !evdref ~pb:Reduction.CONV t t' with
+  match Reductionops.infer_conv env !evdref ~pb:Conversion.CONV t t' with
   | Some sigma -> (evdref := sigma; true)
   | None -> false
-  | exception Reduction.NotConvertible -> false
+  | exception Conversion.NotConvertible -> false
       
 let deps_of_var sigma id env =
   Environ.fold_named_context
@@ -897,8 +924,8 @@ let named_of_rel_context ?(keeplets = false) default l =
     List.fold_right
       (fun decl (subst, args, ids, ctx) ->
         let decl = Context.Rel.Declaration.map_constr (substl subst) decl in
-	let id = match get_name decl with Anonymous -> default () | Name id -> id in
-        let d = Named.Declaration.of_tuple (annotR id, get_value decl, get_type decl) in
+    let id = match get_name decl with Anonymous -> default () | Name id -> id in
+    let d = Named.Declaration.of_rel_decl (fun _ -> id) decl in
 	let args = if keeplets ||Context.Rel.Declaration.is_local_assum decl then mkVar id :: args else args in
 	  (mkVar id :: subst, args, id :: ids, d :: ctx))
       l ([], [], [], [])
@@ -1005,7 +1032,7 @@ let hintdb_set_transparency cst b db =
     (Hints.HintsTransparencyEntry (Hints.HintsReferences [Tacred.EvalConstRef cst], b))
 
 (* Call the really unsafe is_global test, we use this on evar-open terms too *)
-let is_global sigma f ec = EConstr.isRefX sigma f ec
+let is_global = EConstr.isRefX
 
 let constr_of_global_univ sigma u = of_constr (Constr.mkRef (from_peuniverses sigma u))
 
@@ -1068,19 +1095,19 @@ let evd_comb1 f evd x =
 (* Universe related functions *)
 
 let nonalgebraic_universe_level_of_universe env sigma u =
-  match u with
+  match ESorts.kind sigma u with
   | Sorts.Set | Sorts.Prop | Sorts.SProp ->
     sigma, Univ.Level.set, u
-  | Sorts.Type u0 ->
+  | Sorts.Type u0 | Sorts.QSort (_, u0) ->
     match Univ.Universe.level u0 with
     | Some l ->
       (match Evd.universe_rigidity sigma l with
       | Evd.UnivFlexible true ->
-        Evd.make_nonalgebraic_variable sigma l, l, Sorts.sort_of_univ @@ Univ.Universe.make l
+        Evd.make_nonalgebraic_variable sigma l, l, ESorts.make @@ Sorts.sort_of_univ @@ Univ.Universe.make l
       | _ -> sigma, l, u)
     | None ->
       let sigma, l = Evd.new_univ_level_variable Evd.univ_flexible sigma in
-      let ul = Sorts.sort_of_univ @@ Univ.Universe.make l in
+      let ul = ESorts.make @@ Sorts.sort_of_univ @@ Univ.Universe.make l in
       let sigma = Evd.set_leq_sort env sigma u ul in
       sigma, l, ul
 
