@@ -36,10 +36,10 @@ let mkcase env sigma c ty constrs =
     Array.map2_i (fun i id (ctx, cty) ->
       let cty = Term.it_mkProd_or_LetIn cty ctx in
       let cty = CVars.subst_instance_constr ui cty in
-      let (args, arity) = decompose_prod_assum sigma (of_constr cty) in
+      let (args, arity) = decompose_prod_decls sigma (of_constr cty) in
       let realargs, pars = List.chop (List.length args - params) args in
       let args = substl (List.rev origparams) (it_mkProd_or_LetIn arity realargs) in
-      let args, arity = decompose_prod_assum sigma args in
+      let args, arity = decompose_prod_decls sigma args in
       let res = constrs ind i id params args arity in
       it_mkLambda_or_LetIn res args)
       oneind.mind_consnames oneind.mind_nf_lc
@@ -67,9 +67,9 @@ let derive_no_confusion ~pm env sigma0 ~poly (ind,u as indu) =
   let args = oneind.mind_nrealargs in
   let argsvect = rel_vect 0 len in
   let paramsvect, rest = Array.chop params argsvect in
-  let argty, x, ctx, argsctx =
+  let argr, argty, x, ctx, argsctx =
     if Array.length rest = 0 then
-      mkApp (mkIndU indu, argsvect), mkRel 1, ctx, []
+      oneind.mind_relevance, mkApp (mkIndU indu, argsvect), mkRel 1, ctx, []
     else
       let evm, pred, pars, indty, valsig, ctx, lenargs, idx =
         Sigma_types.build_sig_of_ind env !evd indu
@@ -77,25 +77,25 @@ let derive_no_confusion ~pm env sigma0 ~poly (ind,u as indu) =
       let () = evd := evm in
       let evm, sigma = Evd.fresh_global (Global.env ()) !evd (Lazy.force coq_sigma) in
       let () = evd := evm in
-      let _, pred' = Term.decompose_lam_n (List.length pars) (EConstr.to_constr !evd pred) in
+      let _, pred' = Term.decompose_lambda_n (List.length pars) (EConstr.to_constr !evd pred) in
       let indty = mkApp (sigma, [|idx; of_constr pred'|]) in
-      nf_betaiotazeta env !evd indty, mkProj (Lazy.force coq_pr2, mkRel 1), pars, (List.firstn lenargs ctx)
+      Sorts.Relevant, nf_betaiotazeta env !evd indty, mkProj (Lazy.force coq_pr2, mkRel 1), pars, (List.firstn lenargs ctx)
   in
   let tru = get_efresh logic_top evd in
   let fls = get_efresh logic_bot evd in
   let xid = Id.of_string "x" and yid = Id.of_string "y" in
-  let xdecl = of_tuple (nameR xid, None, argty) in
+  let xdecl = of_tuple (make_annot (Name xid) argr, None, argty) in
   let binders = xdecl :: ctx in
-  let ydecl = of_tuple (nameR yid, None, lift 1 argty) in
+  let ydecl = of_tuple (make_annot (Name yid) argr, None, lift 1 argty) in
   let fullbinders = ydecl :: binders in
   let s = Lazy.force logic_sort in
   let s = match s with
     | Sorts.InSProp -> mkSProp
     | Sorts.InProp -> mkProp
     | Sorts.InSet -> mkSet
-    | Sorts.InType ->
+    | Sorts.InType | Sorts.InQSort ->
       (* In that case the noConfusion principle lives at the level of the type. *)
-      let sort = EConstr.mkSort inds in
+      let sort = EConstr.mkSort (ESorts.make inds) in
       let sigma, s =
         Evarsolve.refresh_universes ~status:Evd.univ_flexible ~onlyalg:true
           (Some false) env !evd sort
@@ -114,7 +114,7 @@ let derive_no_confusion ~pm env sigma0 ~poly (ind,u as indu) =
       (* In pars ; x |- fun args (x : ind pars args) => forall y, Prop *)
       let app = pack_ind_with_parlift (args + 2) in
 	      it_mkLambda_or_LetIn 
-          (mkProd_or_LetIn (of_tuple (anonR, None, app)) s)
+          (mkProd_or_LetIn (of_tuple (make_annot Anonymous argr, None, app)) s)
           (of_tuple (nameR xid, None, ind_with_parlift (lenindices + 1)) ::
              lift_rel_context 1 argsctx)
     in
