@@ -118,7 +118,7 @@ let clean_rec_calls sigma (hyps, c) =
   in
   let hyps =
     CMap.fold (fun ty n hyps ->
-      let ctx, concl = Term.decompose_prod_assum ty in
+      let ctx, concl = Term.decompose_prod_decls ty in
       let len = List.length ctx in
       if Vars.noccur_between 1 len concl then
           if CMap.mem (Constr.lift (-len) concl) hyps then hyps
@@ -142,7 +142,7 @@ let rec is_applied_to_structarg f is_rec lenargs =
   | Some (Guarded ids) :: rest -> begin
      try
        let kind =
-         CList.find_map (fun (f', k) -> if Id.equal f f' then Some k else None) ids
+         CList.find_map_exn (fun (f', k) -> if Id.equal f f' then Some k else None) ids
        in
        match kind with
        | MutualOn (Some (idx,_)) | NestedOn (Some (idx,_)) -> Some (lenargs > idx)
@@ -224,7 +224,7 @@ let find_rec_call is_rec sigma protos f args =
       match alias with
       | Some (f',argsf) ->
         let signlen = List.length sign in        
-        let f', args' = Termops.decompose_app_vect sigma f' in
+        let f', args' = EConstr.decompose_app sigma f' in
         let f' = EConstr.Unsafe.to_constr f' in
         if Constr.equal (head f') f then
           let sign, args = 
@@ -246,8 +246,7 @@ let find_rec_call is_rec sigma protos f args =
         else None
       | None -> None
   in
-  try Some (CList.find_map fm protos)
-  with Not_found -> None
+  CList.find_map fm protos
 
 let filter_arg i filter =
   let rec aux f =
@@ -293,7 +292,7 @@ let abstract_rec_calls sigma user_obls ?(do_subst=true) is_rec len protos c =
         hyps', mkProj (p, c')
 
     | _ ->
-      let f', args = decompose_appvect c in
+      let f', args = decompose_app c in
       if not (is_user_obl sigma user_obls (EConstr.of_constr f')) then
         (match find_rec_call is_rec sigma protos f' (Array.to_list args) with
          | Some (i, arity, filter, sign, (fargs', indargs', rest)) ->
@@ -354,7 +353,7 @@ let subst_comp_proj_split sigma f proj s =
   map_split (subst_comp_proj sigma f proj) s
 
 let is_ind_assum env sigma ind b =
-  let _, concl = decompose_prod_assum sigma b in
+  let _, concl = decompose_prod_decls sigma b in
   let t, _ = decompose_app sigma concl in
   if isInd sigma t then
     let (ind', _), _ = destInd sigma t in
@@ -381,7 +380,7 @@ open Vars
 
 let compute_elim_type env evd user_obls is_rec protos k leninds
                       ind_stmts all_stmts sign app elimty =
-  let ctx, arity = decompose_prod_assum !evd elimty in
+  let ctx, arity = decompose_prod_decls !evd elimty in
   let lenrealinds =
     List.length (List.filter (fun (_, (_,_,_,_,_,_,_,(kind,_)),_) -> regular_or_nested_rec kind) ind_stmts) in
   let newctx =
@@ -394,7 +393,7 @@ let compute_elim_type env evd user_obls is_rec protos k leninds
       it_mkProd_or_LetIn (Vars.substl [mkProp; app] arity) sign
     else
       let clean_one a sign fn =
-        let ctx, concl = decompose_prod_assum !evd a in
+        let ctx, concl = decompose_prod_decls !evd a in
         let newctx = CList.skipn 2 ctx in
         let newconcl = Vars.substl [mkProp; mkApp (fn, extended_rel_vect 0 sign)] concl in
         it_mkProd_or_LetIn newconcl newctx
@@ -610,7 +609,7 @@ let pr_where env sigma ctx ({where_type} as w) =
 let where_instance w =
   List.map (fun w -> where_term w) w
 
-let arguments sigma c = snd (Termops.decompose_app_vect sigma c)
+let arguments sigma c = snd (decompose_app sigma c)
 
 let unfold_constr sigma c =
   Tactics.unfold_in_concl [(Locus.OnlyOccurrences [1], Tacred.EvalConstRef (fst (destConst sigma c)))]
@@ -621,7 +620,7 @@ let extend_prob_ctx delta (ctx, pats, ctx') =
 let map_proto evd recarg f ty =
   match recarg with
   | Some recarg ->
-     let lctx, ty' = decompose_prod_assum evd ty in
+     let lctx, ty' = decompose_prod_decls evd ty in
      (* Feedback.msg_debug Pp.(str"map_proto: " ++ Printer.pr_econstr_env (Global.env()) evd ty ++ str" recarg = " ++ int recarg); *)
      let app =
        let args = Termops.rel_list 0 (List.length lctx) in
@@ -816,7 +815,7 @@ let subst_rec_programs env evd ps =
       let splitting' =
           aux rec_cutprob s' program' oterm path' p.program_splitting
       in
-      let term', ty' = term_of_tree env evd prog_info.program_sort splitting' in
+      let term', ty' = term_of_tree env evd (ESorts.make prog_info.program_sort) splitting' in
       { program_rec = None;
         program_info = program_info';
         program_prob = id_subst (pi3 cutprob_sign);
@@ -940,7 +939,7 @@ let subst_rec_programs env evd ps =
       let refarg = ref (0,0) in
       let refhead =
         if islogical then
-          let term', _ = term_of_tree env evd p.program_info.program_sort s' in
+          let term', _ = term_of_tree env evd (ESorts.make p.program_info.program_sort) s' in
           term'
          else mapping_constr !evd subst oterm
       in
@@ -1113,7 +1112,7 @@ let computations env evd alias refine p eqninfo : computation list =
      let where_comp w (wheres, where_comps) =
        (* Where term is in lhs + wheres *)
        let lhsterm = substl wheres (where_term w) in
-       let term, args = decompose_app evd lhsterm in
+       let term, args = decompose_app_list evd lhsterm in
        let alias, fsubst =
          try
            let (f, id, s) = PathMap.find w.where_path wheremap in
@@ -1279,7 +1278,8 @@ let declare_funelim ~pm info env evd is_rec protos progs
                            Himsg.explain_pretype_error env !evd
                              (Pretype_errors.TypingError (Type_errors.map_ptype_error EConstr.of_constr tyerr)))
   in
-  let cinfo = Declare.CInfo.make ~name:(Nameops.add_suffix id "_elim") ~typ:(to_constr !evd newty) () in
+  let newty = collapse_term_qualities (Evd.evar_universe_context !evd) (EConstr.to_constr !evd newty) in 
+  let cinfo = Declare.CInfo.make ~name:(Nameops.add_suffix id "_elim") ~typ:newty () in
   let info = Declare.Info.make ~poly:info.poly ~scope:info.scope ~hook:(Declare.Hook.make hookelim) ~kind:(Decls.IsDefinition info.decl_kind) () in
   let pm, _ = Declare.Obls.add_definition ~pm ~cinfo ~info ~tactic
       ~uctx:(Evd.evar_universe_context !evd) [||] in
@@ -1290,7 +1290,7 @@ let mkConj evd sort x y =
     mkApp (prod, [| x; y |])
 
 let declare_funind ~pm info alias env evd is_rec protos progs
-                   ind_stmts all_stmts sign inds kn comb sort f split ind =
+                   ind_stmts all_stmts sign inds kn comb sort f split =
   let poly = is_polymorphic info.term_info in
   let id = Id.of_string info.term_info.base_id in
   let indid = Nameops.add_suffix id "_graph_correct" in
@@ -1386,7 +1386,8 @@ let declare_funind ~pm info alias env evd is_rec protos progs
   in
   let evm, stmtt = Typing.type_of (Global.env ()) !evd statement in
   let () = evd := evm in
-  let stmt = to_constr !evd statement and f = to_constr !evd f in
+  let to_constr c = collapse_term_qualities (Evd.evar_universe_context !evd) (EConstr.to_constr !evd c) in
+  let stmt = to_constr statement and f = to_constr f in
   let uctx = Evd.evar_universe_context (if poly then !evd else Evd.from_env (Global.env ())) in
   let launch_ind ~pm tactic =
     let pm, res =
@@ -1411,7 +1412,7 @@ let declare_funind ~pm info alias env evd is_rec protos progs
     CErrors.user_err Pp.(str"Functional induction principle could not be proved automatically: " ++
                          Himsg.explain_pretype_error env !evd
                            (Pretype_errors.TypingError (Type_errors.map_ptype_error EConstr.of_constr tyerr)))
-     | e ->
+     | e when CErrors.noncritical e ->
        Feedback.msg_warning Pp.(str "Functional induction principle could not be proved automatically: " ++ fnl () ++
                                 CErrors.print e);
        launch_ind ~pm (Proofview.tclUNIT ())
@@ -1424,11 +1425,13 @@ let max_sort s1 s2 =
   | (Prop, (Set | Type _ as s)) | ((Set | Type _) as s, Prop) -> s
   | (Set, Type u) | (Type u, Set) -> Sorts.sort_of_univ (Univ.Universe.sup Univ.Universe.type0 u)
   | (Type u, Type v) -> Sorts.sort_of_univ (Univ.Universe.sup u v)
+  | (QSort _, _) | (_, QSort _) -> assert false
 
 let level_of_context env evd ctx acc =
   let _, lev =
     List.fold_right (fun decl (env, lev) ->
         let s = Retyping.get_sort_of env evd (get_type decl) in
+        let s = ESorts.kind evd s in
         (push_rel decl env, max_sort s lev))
                     ctx (env,acc)
   in lev
@@ -1532,7 +1535,7 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
     CList.map_i (fun i ((f',filterf'), alias, path, sign, arity, pats, args, (refine, cut)) ->
       let f' = Termops.strip_outer_cast evd f' in
       let f'hd =
-        let ctx, t = decompose_lam_assum evd f' in
+        let ctx, t = decompose_lambda_decls evd f' in
         fst (decompose_app evd t)
       in
       let alias =
@@ -1585,7 +1588,7 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
             let f = mkRel (len + (lenprotos - i) + hypslen) in
             if cut then f
             else
-              let fn, args = decompose_app !evd (Termops.strip_outer_cast !evd fl) in
+              let fn, args = decompose_app_list !evd (Termops.strip_outer_cast !evd fl) in
               applistc f (filter_arguments filter (lift_constrs hypslen args))
           in
           let ty =
@@ -1645,7 +1648,7 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
     let entry =
       Entries.{ mind_entry_typename = indid;
                 mind_entry_arity = to_constr !evd (it_mkProd_or_LetIn (mkProd (anonR, arity,
-                                                                               mkSort ind_sort)) sign);
+                                                                               mkSort (ESorts.make ind_sort))) sign);
                 mind_entry_consnames = consnames;
                 mind_entry_lc = constructors;
               }
@@ -1655,13 +1658,17 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
     let inds, univs, sort = List.fold_left declare_one_ind ([], Univ.Level.Set.empty, Sorts.prop) ind_stmts in
     let sigma = Evd.restrict_universe_context !evd univs in
     let sigma = Evd.minimize_universes sigma in
+    (* FIXME: try to implement a sane handling of universe state threading *)
+    let to_constr sigma c =
+      collapse_term_qualities (Evd.evar_universe_context sigma) (EConstr.to_constr sigma c)
+    in
     let inds =
       List.rev_map (fun (entry, sign, arity) ->
           Entries.{ entry with
                     mind_entry_lc = List.map (to_constr sigma) (List.map of_constr entry.mind_entry_lc);
                     mind_entry_arity =
                       to_constr sigma (it_mkProd_or_LetIn (mkProd (anonR, arity,
-                                                                  mkSort sort)) sign) })
+                                                                  mkSort (ESorts.make sort))) sign) })
         inds
     in
     let univs, ubinders = Evd.univ_entry ~poly sigma in
@@ -1710,18 +1717,15 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
                          i, regular_or_nested_rec kind) mutual ind_stmts in
         let () =
           Indschemes.do_combined_scheme CAst.(make scheme)
-            (CList.map_filter (fun (id, b) -> if b then Some id else None) mutual)
+            (CList.map_filter (fun ({CAst.loc;v=id}, b) ->
+                 if b
+                 then Some (Nametab.locate_constant (Libnames.qualid_of_ident ?loc id))
+                 else None)
+                mutual)
         in kn, Smartlocate.global_with_alias (Libnames.qualid_of_ident scheme)
     in
-    let ind =
-      let open Entries in
-      match uctx with
-      | Polymorphic_ind_entry uctx ->
-        mkIndU ((kn,0), EInstance.make (Univ.UContext.instance uctx))
-      | Monomorphic_ind_entry | Template_ind_entry _ -> mkInd (kn,0)
-    in
     let locality = if Global.sections_are_opened () then Hints.Local else Hints.SuperGlobal in
-    let _ =
+    let () =
       List.iteri (fun i ind ->
         let constrs =
           CList.map_i (fun j _ -> Hints.empty_hint_info, true, Hints.PathAny,
@@ -1732,7 +1736,7 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
     let info = { term_info = info; pathmap = !fnind_map; wheremap } in
     declare_funind ~pm info alias (Global.env ()) evd rec_info protos progs
                    ind_stmts all_stmts sign inds kn comb sort
-                   f p.program_splitting ind
+                   f p.program_splitting
   in
   let () = evd := Evd.minimize_universes !evd in
   let () =
@@ -1788,7 +1792,9 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
         if not poly then evd := Evd.from_env (Global.env ())
         else ()
       in
-      let cinfo = Declare.CInfo.make ~name:ideq ~typ:(to_constr !evd c) () in
+      (* FIXME: try to implement a sane handling of universe state threading *)
+      let c = collapse_term_qualities (Evd.evar_universe_context !evd) (to_constr !evd c) in
+      let cinfo = Declare.CInfo.make ~name:ideq ~typ:c () in
       let info = Declare.Info.make ~kind:(Decls.IsDefinition info.decl_kind) ~poly () in
       let obl_hook = Declare.Hook.make_g hook in
       let pm, _ = Declare.Obls.add_definition ~pm ~cinfo ~info ~obl_hook
