@@ -44,10 +44,6 @@ val destWheres : Splitting.splitting -> (Context_map.context_map * Splitting.whe
 val map_opt_split : ('a -> 'b option) -> 'a option -> 'b option
 val solve_ind_rec_tac : Splitting.term_info -> unit Proofview.tactic
 
-val gather_subst :            'a ->
-           Evd.evar_map ->
-           EConstr.t -> EConstr.t list -> int -> EConstr.t list
-
 val aux_ind_fun :
   ind_info ->
   int * int ->
@@ -63,6 +59,12 @@ val ind_fun_tac :
   (Splitting.program * Splitting.program option * Splitting.compiled_program_info * 'b) list ->
   unit Proofview.tactic
 
+type unfold_subst
+
+val extract_subprograms :
+  Environ.env -> Evd.evar_map -> where_map -> Splitting.program -> Splitting.program ->
+    (Names.Id.t * unfold_subst * UState.t * EConstr.t) list
+
 val prove_unfolding_lemma :
   Splitting.term_info ->
   where_map ->
@@ -71,6 +73,14 @@ val prove_unfolding_lemma :
   Splitting.program -> Splitting.program ->
   unit Proofview.tactic
   
+val prove_unfolding_sublemma :
+  Splitting.term_info ->
+  where_map ->
+  Names.Constant.t ->
+  Names.Constant.t ->
+  unfold_subst ->
+  unit Proofview.tactic
+
 val ind_elim_tac :
   constr ->
   int -> int ->
