@@ -29,21 +29,21 @@ let mkConstructG c u =
 let mkIndG c u =
   mkIndU (destIndRef (Lazy.force c), u)
 
-let mkAppG evd gr args = 
+let mkAppG env evd gr args =
   let c = e_new_global evd gr in
-    mkApp (c, args)
+  evd_comb1 (fun sigma c -> Typing.checked_appvect env sigma c args) evd c
 
-let applistG evd gr args = 
-  mkAppG evd gr (Array.of_list args)
+let applistG env evd gr args =
+  mkAppG env evd gr (Array.of_list args)
 
-let mkSig evd (n, c, t) = 
+let mkSig env evd (n, c, t) =
   let args = [| c; mkLambda (n, c, t) |] in
-    mkAppG evd (Lazy.force coq_sigma) args
+    mkAppG env evd (Lazy.force coq_sigma) args
 
 let constrs_of_coq_sigma env evd t alias = 
   let rec aux env proj c ty =
     match kind !evd c with
-    | App (f, args) when is_global !evd (Lazy.force coq_sigmaI) f && 
+    | App (f, args) when is_global env !evd (Lazy.force coq_sigmaI) f && 
 	                   Array.length args = 4 ->
        let ty = Retyping.get_type_of env !evd args.(1) in
 	(match kind !evd ty with
@@ -56,10 +56,10 @@ let constrs_of_coq_sigma env evd t alias =
     | _ -> [(Context.anonR, c, proj, ty)]
   in aux env alias t (Retyping.get_type_of env !evd t)
 
-let decompose_coq_sigma sigma t = 
+let decompose_coq_sigma env sigma t = 
   let s = Lazy.force coq_sigma in
     match kind sigma t with
-    | App (f, args) when is_global sigma s f && Array.length args = 2 ->
+    | App (f, args) when is_global env sigma s f && Array.length args = 2 ->
        let ind, u = destInd sigma f in
          Some (u, args.(0), args.(1))
     | _ -> None
@@ -113,6 +113,11 @@ let telescope_of_context env sigma ctx =
   let sigma, _ = Typing.type_of env sigma tele_interp in
   sigma, tele, tele_interp
 
+(* Given a context ⊢ Γ, returns
+  - ⊢ Tel(Γ) : Type (iterated Σ-type built out of Γ)
+  - p : Tel(Γ) ⊢ Ctx(Γ) (a context of let-bindings made of each projection in order)
+  - Γ ⊢ Intro(Γ) : Tel(Γ) (iterated constructors)
+*)
 let telescope env evd = function
   | [] -> assert false
   | [d] -> let (n, _, t) = to_tuple d in t, [of_tuple (n, Some (mkRel 1), Vars.lift 1 t)],
@@ -121,6 +126,7 @@ let telescope env evd = function
       let (n, _, t) = to_tuple d in
       let len = succ (List.length tl) in
       let ts = Retyping.get_sort_of (push_rel_context tl env) !evd t in
+      let ts = ESorts.kind !evd ts in
       let ty, tys =
         let rec aux (ty, tyuniv, tys) ds =
           match ds with
@@ -128,15 +134,16 @@ let telescope env evd = function
           | d :: ds ->
             let (n, b, t) = to_tuple d in
             let pred = mkLambda (n, t, ty) in
-            let sigty = mkAppG evd (Lazy.force coq_sigma) [|t; pred|] in
+            let env = push_rel_context ds env in
+            let sigty = mkAppG env evd (Lazy.force coq_sigma) [|t; pred|] in
             let _, u = destInd !evd (fst (destApp !evd sigty)) in
             let ua = Univ.Instance.to_array (EInstance.kind !evd u) in
             let l = Sorts.sort_of_univ @@ Univ.Universe.make ua.(0) in
-            let env = push_rel_context ds env in
             (* Ensure that the universe of the sigma is only >= those of t and pred *)
             let open UnivProblem in
             let enforce_leq env sigma t cstr =
               let ts = Retyping.get_sort_of env sigma t in
+              let ts = ESorts.kind sigma ts in
               UnivProblem.Set.add (ULe (ts, l)) cstr
             in
             let cstrs = enforce_leq env !evd t (UnivProblem.Set.add (ULe (tyuniv, l)) UnivProblem.Set.empty) in
@@ -166,7 +173,7 @@ let telescope env evd = function
 let sigmaize ?(liftty=0) env0 evd pars f =
   let env = push_rel_context pars env0 in
   let ty = Retyping.get_type_of env !evd f in
-  let ctx, concl = splay_prod_assum env !evd ty in
+  let ctx, concl = hnf_decompose_prod_decls env !evd ty in
   let ctx = EConstr.Vars.smash_rel_context ctx in
   let argtyp, letbinders, make = telescope env evd ctx in
     (* Everyting is in env, move to index :: letbinders :: env *) 
@@ -179,7 +186,7 @@ let sigmaize ?(liftty=0) env0 evd pars f =
 	        letbinders)
   in
   let tyargs = [| argtyp; pred |] in
-  let tysig = mkAppG evd (Lazy.force coq_sigma) tyargs in
+  let tysig = mkAppG env evd (Lazy.force coq_sigma) tyargs in
   let indexproj = Lazy.force coq_pr1 in
   let valproj = Lazy.force coq_pr2 in
   let indices = 
@@ -195,9 +202,8 @@ let sigmaize ?(liftty=0) env0 evd pars f =
 			rel_vect 0 lenb))
 		 (Equations_common.lift_rel_contextn 1 (succ lenb) letbinders))
     in
-    let _tylift = Vars.lift lenb argtyp in
-      mkAppG evd (Lazy.force coq_sigmaI)
-	[|argtyp; pred; Vars.lift 1 make; mkRel 1|]
+    let (_, u) = destInd !evd (fst @@ destApp !evd tysig) in
+    mkApp (mkRef (Lazy.force coq_sigmaI, u), [|argtyp; pred; Vars.lift 1 make; mkRel 1|])
   in
   let pred = it_mkLambda_or_LetIn pred pars in
   let _ = e_type_of env0 evd pred in
@@ -368,7 +374,7 @@ let curry_left_hyp env sigma c t =
   let rec curry_index c t =
     match kind sigma t with
     | Prod (na, dom, concl) ->
-       (match decompose_coq_sigma sigma dom with
+       (match decompose_coq_sigma env sigma dom with
 	| None -> (c, t)
 	| Some (u, ty, pred) ->
 	   let term, typ = aux c t na u ty pred concl in
@@ -383,7 +389,7 @@ let curry_left_hyp env sigma c t =
   let curry c t =
     match kind sigma t with
     | Prod (na, dom, concl) ->
-       (match decompose_coq_sigma sigma dom with
+       (match decompose_coq_sigma env sigma dom with
 	| None -> None
 	| Some (inst, ty, pred) ->
 	   let term, typ = aux c t na inst ty pred concl in
@@ -392,11 +398,11 @@ let curry_left_hyp env sigma c t =
     | _ -> None
   in curry c t
 
-let curry sigma na c =
+let curry env sigma na c =
   let rec make_arg na t =
-    match decompose_coq_sigma sigma t with
+    match decompose_coq_sigma env sigma t with
     | None -> 
-       if is_global sigma (Lazy.force logic_unit) t then
+       if is_global env sigma (Lazy.force logic_unit) t then
          let _, u = destInd sigma t in
          [], constr_of_global_univ sigma (Lazy.force logic_unit_intro, u)
        else [of_tuple (na,None,t)], mkRel 1
@@ -423,7 +429,7 @@ let uncurry_hyps name =
     let sigma = Goal.sigma gl in
     let hyps, _ =
       List.split_when (fun d ->
-          is_global sigma (Lazy.force coq_end_of_section) (get_named_type d)
+          is_global env sigma (Lazy.force coq_end_of_section) (get_named_type d)
           || is_section_variable (Global.env ()) (get_id d)) hyps in
     let ondecl (sigma, acc, ty) d =
       let (dna, _, dty) = to_named_tuple d in
@@ -446,10 +452,10 @@ let uncurry_hyps name =
 let uncurry_call env sigma fn c =
   let hd', args' = decompose_app sigma fn in
   let hd, args = decompose_app sigma c in
-  let params, args = List.chop (List.length args') args in
-  let hd = applist (hd, params) in
+  let params, args = Array.chop (Array.length args') args in
+  let hd = mkApp (hd, params) in
   let ty = Retyping.get_type_of env sigma hd in
-  let ctx, _ = Reductionops.splay_prod_assum env sigma ty in
+  let ctx, _ = Reductionops.hnf_decompose_prod_decls env sigma ty in
   let ctx =
     let open Context.Rel.Declaration in
     let rec aux env ctx args =
@@ -463,14 +469,14 @@ let uncurry_call env sigma fn c =
         decl :: aux env decls args
       | [], _ :: _ -> assert false
       | ctx, [] -> []
-    in List.rev (aux env (List.rev ctx) args)
+    in List.rev (aux env (List.rev ctx) (Array.to_list args))
   in
   let evdref = ref sigma in
   if CList.is_empty ctx then 
     user_err_loc (None, Pp.str"No arguments to uncurry");
   (* let ctx = (Anonymous, None, concl) :: ctx in *)
   let sigty, sigctx, constr = telescope env evdref ctx in
-  let app = Vars.substl (List.rev args) constr in
+  let app = Vars.substl (Array.rev_to_list args) constr in
   let fnapp = mkApp (hd, rel_vect 0 (List.length sigctx)) in
   let fnapp = it_mkLambda_or_subst env fnapp sigctx in
   let projsid = nameR (Id.of_string "projs") in
@@ -762,7 +768,7 @@ let smart_case (env : Environ.env) (evd : Evd.evar_map ref)
     let args = List.map of_rel_decl args in
     (* Substitute the indices in [cuts_ctx]. *)
     let rev_indices = List.rev indices in
-    let pats_indices = List.map (Context_map.pat_of_constr !evd) rev_indices in
+    let pats_indices = List.map (Context_map.pat_of_constr env !evd) rev_indices in
     let pats_ctx' = Context_map.lift_pats summary.Inductiveops.cs_nargs pats_ctx' in
     let pats = pats_indices @ pats_ctx' in
     let cuts_ctx = Context_map.specialize_rel_context !evd pats cuts_ctx in
@@ -794,7 +800,7 @@ let curry_hyp env sigma hyp t =
   let curry t =
     match kind sigma t with
     | Prod (na, dom, concl) ->
-       let ctx, arg = curry sigma na dom in
+       let ctx, arg = curry env sigma na dom in
        let term = mkApp (mkVar hyp, [| arg |]) in
        let ty = Reductionops.nf_betaiota env sigma (Vars.subst1 arg concl) in
        Some (it_mkLambda_or_LetIn term ctx, it_mkProd_or_LetIn ty ctx)
@@ -812,7 +818,7 @@ let red_curry () =
   Reductionops.clos_norm_flags reds
 
 let curry_concl env sigma na dom codom =
-  let ctx, arg = curry sigma na dom in
+  let ctx, arg = curry env sigma na dom in
   let newconcl =
     let body = it_mkLambda_or_LetIn (Vars.subst1 arg codom) ctx in
     let inst = extended_rel_vect 0 ctx in
