@@ -176,7 +176,7 @@ let pr_program_info env sigma p =
   let open Pp in
   Names.Id.print p.program_id ++ str " : " ++
   Printer.pr_econstr_env env sigma (Syntax.program_type p) ++ str " : " ++
-  Printer.pr_econstr_env env sigma (mkSort p.program_sort) ++
+  Printer.pr_econstr_env env sigma (mkSort (ESorts.make p.program_sort)) ++
   str " ( " ++
   (match p.program_rec with
    | Some (Structural ann) ->
@@ -348,7 +348,7 @@ let compute_possible_guardness_evidences sigma n fixbody fixtype =
          doesn't seem to worth the effort (except for huge mutual
          fixpoints ?) *)
     let m = Termops.nb_prod sigma fixtype in
-    let ctx = fst (decompose_prod_n_assum sigma m fixtype) in
+    let ctx = fst (decompose_prod_n_decls sigma m fixtype) in
     List.map_i (fun i _ -> i) 0 ctx
 
 let nf_fix sigma (nas, cs, ts) =
@@ -375,7 +375,7 @@ let define_mutual_nested env evd get_prog progs =
       in
       let after = (nested - 1) - before in
       let fixb = (Array.make 1 idx, 0) in
-      let fixna = Array.make 1 (nameR p.program_id) in
+      let fixna = Array.make 1 (make_annot (Name p.program_id) (Retyping.relevance_of_sort !evd (ESorts.make p.program_sort))) in
       let fixty = Array.make 1 (Syntax.program_type p) in
       (* Apply to itself *)
       let beforeargs = Termops.rel_list (signlen + 1) before in
@@ -471,6 +471,23 @@ let define_mutual_nested env evd get_prog progs =
   let nested = List.map2 declare_nested nested nestedbodies in
   fixes, nested
 
+let check_typed ~where ?name env evd c =
+  let sigma, _ =
+    try Typing.type_of env evd c
+    with Type_errors.TypeError (env, tyerr) ->
+      anomaly Pp.(str where ++ spc () ++
+        str "Equations build an ill-typed term: " ++ Printer.pr_econstr_env env evd c ++
+        Himsg.explain_pretype_error env evd
+          (Pretype_errors.TypingError (Type_errors.map_ptype_error EConstr.of_constr tyerr)))
+    | Pretype_errors.PretypeError (env, evd, tyerr) ->
+        anomaly Pp.(str where ++ spc () ++
+        str "Equations build an ill-typed term: " ++ Printer.pr_econstr_env env evd c ++
+        Himsg.explain_pretype_error env evd tyerr)
+  in
+  let check = Evd.check_constraints evd (snd @@ Evd.universe_context_set sigma) in
+  if not check then anomaly Pp.(str where ++ spc () ++ str "Equations missing constraints in " ++
+    str (Option.default "(anonymous)" name))
+
 let term_of_tree env0 isevar sort tree =
   let rec aux env evm sort = function
     | Compute ((ctx, _, _), where, ty, RProgram rhs) ->
@@ -529,7 +546,7 @@ let term_of_tree env0 isevar sort tree =
       in
       let branches = Array.map2 (fun (ty, nb, csubst) next ->
         (* We get the context from the constructor arity. *)
-        let new_ctx, ty = EConstr.decompose_prod_n_assum !isevar nb ty in
+        let new_ctx, ty = EConstr.decompose_prod_n_decls !isevar nb ty in
         let new_ctx = Namegen.name_context env !isevar new_ctx in
         let envnew = push_rel_context (new_ctx @ ctx') env in
         (* Remove the cuts and append them to the context. *)
@@ -618,7 +635,7 @@ let term_of_tree env0 isevar sort tree =
       (* Remove the block lets *)
       let rec clean_block c =
         match kind !evd c with
-        | LetIn (_, b, _, b') when Equations_common.is_global !evd (Lazy.force coq_block) b ->
+        | LetIn (_, b, _, b') when Equations_common.is_global env !evd (Lazy.force coq_block) b ->
           clean_block (subst1 b b')
         | _ -> EConstr.map !evd clean_block c
       in
@@ -640,8 +657,8 @@ let term_of_tree env0 isevar sort tree =
       let term = EConstr.mkApp (case, Array.of_list to_apply) in
       let term = EConstr.it_mkLambda_or_LetIn term ctx in
       let typ = it_mkProd_or_subst env evm ty ctx in
+      let () = if !Equations_common.debug then check_typed ~where:"splitting" env !evd term in
       let term = Evarutil.nf_evar !evd term in
-      evd := Typing.check env !evd term typ;
       !evd, term, typ
   in
   let evm, term, typ = aux env0 !isevar sort tree in
@@ -683,7 +700,7 @@ type program_shape =
 let make_program env evd p prob s rec_info =
   match rec_info with
   | Some r ->
-    let sort = p.program_sort in
+    let sort = ESorts.make p.program_sort in
     let lhs = id_subst r.rec_lets in
     (match r.rec_node with
      | WfRec wfr ->
@@ -725,7 +742,7 @@ let make_program env evd p prob s rec_info =
        in
        Mutual (p', prob, r, s, after, (* lift 1  *)term))
   | None ->
-    Single (p, prob, rec_info, s, fst (term_of_tree env evd p.program_sort s))
+    Single (p, prob, rec_info, s, fst (term_of_tree env evd (ESorts.make p.program_sort) s))
 
 let update_rec_info p rec_info =
   match p.Syntax.program_rec, rec_info.rec_node with
@@ -818,7 +835,7 @@ let change_lhs s (l, p, r) =
   let l' =
     List.map
       (function LocalDef ({binder_name=Name id}, b, t) as decl ->
-         (try let b' = List.assoc id s in LocalDef (nameR id, b', t)
+         (try let b' = List.assoc id s in LocalDef (make_annot (Name id) (get_relevance decl), b', t)
           with Not_found -> decl)
               | decl -> decl) l
   in l', p, r
@@ -905,7 +922,7 @@ let define_one_program_constants flags env0 isevar udecl unfold p =
       let evm, s' = aux env evm p.program_splitting in
       let isevar = ref evm in
       let env = Global.env () in
-      let term, ty = term_of_tree env isevar p.program_info.program_sort s' in
+      let term, ty = term_of_tree env isevar (ESorts.make p.program_info.program_sort) s' in
       let (cst, (evm, e)) =
         Equations_common.declare_constant (path_id (Id.of_string "functional" :: path))
           term (Some ty)
@@ -998,7 +1015,7 @@ let is_comp_obl sigma comp hole_kind =
   | Some r ->
       match hole_kind, r with
       | ImplicitArg (GlobRef.ConstRef c, (n, _), _), (loc, id) ->
-        is_rec_call sigma (snd r) (mkConst c)
+        is_rec_call (snd r) c
       | _ -> false
 
 type term_info = {
@@ -1023,10 +1040,12 @@ let warn_complete id =
   str "Equations definition " ++ Id.print id ++ str" is complete and requires no further proofs. " ++
   str "Use the \"Equations\" command to define it."
 
+let equations_cat = CWarnings.create_category ~name:"equations" ()
+
 let warn_complete = 
   CWarnings.create
     ~name:"equations-open-proof-complete"
-    ~category:"equations"
+    ~category:equations_cat
     ~default:CWarnings.Enabled
     warn_complete
 
@@ -1170,7 +1189,7 @@ let solve_equations_obligations_program ~pm flags recids loc i sigma hook =
     let recobls =
       List.map (fun (id, c) ->
       (* Due to shrinking, we can get lambda abstractions first *)
-      let _, body = decompose_lam_assum sigma (EConstr.of_constr c) in
+      let _, body = decompose_lambda_decls sigma (EConstr.of_constr c) in
       let hd, _ = decompose_app sigma body in
       try fst (EConstr.destConst sigma hd)
       with Constr.DestKO -> assert false) obls
