@@ -25,7 +25,6 @@ open Sigma_types
 let refresh_universes t = t (* MS: FIXME *)
 
 let derive_subterm ~pm env sigma ~poly (ind, u as indu) =
-  let global = true in
   let () = 
     if Ederive.check_derive "NoConfusion" (Names.GlobRef.IndRef ind) 
       || Ederive.check_derive "NoConfusionHom" (Names.GlobRef.IndRef ind) then ()
@@ -46,7 +45,7 @@ let derive_subterm ~pm env sigma ~poly (ind, u as indu) =
     | Sorts.InSProp -> failwith "not implemented"
     | Sorts.InProp -> mkProp
     | Sorts.InSet -> mkSet
-    | Sorts.InType -> EConstr.mkSort indsort
+    | Sorts.InType | Sorts.InQSort -> EConstr.mkSort (ESorts.make indsort)
   in
   let len = List.length ctx in
   let params = mind.mind_nparams_rec in
@@ -54,16 +53,16 @@ let derive_subterm ~pm env sigma ~poly (ind, u as indu) =
   let lenargs = len - params in
   let argbinders, parambinders = List.chop lenargs (List.map of_rel_decl ctx) in
   let indapp = mkApp (mkIndU indu, extended_rel_vect 0 parambinders) in
-  let getargs t = snd (List.chop params (snd (decompose_app sigma t))) in
+  let getargs t = snd (Array.chop params (snd (decompose_app sigma t))) in
   let inds =
     let branches = Array.mapi (fun i ty ->
-      let args, concl = decompose_prod_assum sigma (of_constr ty) in
+      let args, concl = decompose_prod_decls sigma (of_constr ty) in
       let lenargs = List.length args in
       let lenargs' = lenargs - params in
       let args', params' = List.chop lenargs' args in
       let recargs = CList.map_filter_i (fun i decl ->
         let (n, _, t) = to_tuple decl in
-        let ctx, ar = decompose_prod_assum sigma t in
+        let ctx, ar = decompose_prod_decls sigma t in
           match kind sigma (fst (decompose_app sigma ar)) with
           | Ind (ind',_) when Environ.QInd.equal env ind' ind ->
               Some (ctx, i, mkRel (succ i), getargs (lift (succ i) ar))
@@ -77,7 +76,7 @@ let derive_subterm ~pm env sigma ~poly (ind, u as indu) =
           let subargs =
             Array.of_list ((extended_rel_list (lenargs' + ctxlen)
                                parambinders)
-                            @ rargs @ (List.map (lift ctxlen) constrargs) @
+                            @ Array.to_list rargs @ (Array.map_to_list (lift ctxlen) constrargs) @
                             [mkApp (lift ctxlen r, extended_rel_vect 0 ctx) ;
                              lift ctxlen constr])
           in
@@ -92,10 +91,12 @@ let derive_subterm ~pm env sigma ~poly (ind, u as indu) =
   let _trans_branch =
     let liftargbinders = lift_rel_context lenargs argbinders in
     let liftargbinders' = lift_rel_context lenargs liftargbinders in
+    let indr = oneind.mind_relevance in
+    let indna id = make_annot (Name (Id.of_string id)) indr in
     let indapp n = (mkApp (lift (3 * lenargs + n) indapp, extended_rel_vect (n + (2 - n) * lenargs) argbinders)) in
-    let terms = [(nameR (Id.of_string "z"), None, indapp 2);
-                 (nameR (Id.of_string "y"), None, indapp 1);
-                 (nameR (Id.of_string "x"), None, indapp 0)]
+    let terms = [(indna "z", None, indapp 2);
+                 (indna "y", None, indapp 1);
+                 (indna "x", None, indapp 0)]
     in
     let binders = to_context terms @ liftargbinders' @ liftargbinders @ argbinders in
     let lenbinders = 3 * succ lenargs in
@@ -252,8 +253,8 @@ let derive_subterm ~pm env sigma ~poly (ind, u as indu) =
     let body = it_mkLambda_or_LetIn (Option.get body) parambinders in
     let hook { Declare.Hook.S.dref; _ } =
       let cst = match dref with GlobRef.ConstRef kn -> kn | _ -> assert false in
-      Classes.Internal.add_instance (fst kl) empty_hint_info
-                                          global (GlobRef.ConstRef cst)
+      Classes.declare_instance (Global.env ()) !evm (Some empty_hint_info)
+                                          Hints.SuperGlobal (GlobRef.ConstRef cst)
     in
     let _bodyty = e_type_of (Global.env ()) evm body in
     let _ty' = e_type_of (Global.env ()) evm ty in
@@ -284,15 +285,18 @@ let derive_below env sigma ~poly (ind,univ as indu) =
   let ctx = List.map of_rel_decl ctx in
   let allargsvect = extended_rel_vect 0 ctx in
   let indty = mkApp (mkIndU indu, allargsvect) in
-  let ctx = of_tuple (nameR (Id.of_string "c"), None, indty) :: ctx in
+  let indr = oneind.mind_relevance in
+  let ctx = of_tuple (make_annot (Name (Id.of_string "c")) indr, None, indty) :: ctx in
   let argbinders, parambinders = List.chop (succ realdecls) ctx in
-  let u = evd_comb0 (Evarutil.new_Type ~rigid:Evd.univ_rigid) evd in
+  let u = evd_comb0 (Evd.new_sort_variable Evd.univ_rigid) evd in
+  let ru = Retyping.relevance_of_sort !evd u in
+  let u = mkSort u in
   let arity = it_mkProd_or_LetIn u argbinders in
   let aritylam = lift (succ realdecls) (it_mkLambda_or_LetIn u argbinders) in
   let paramsvect = rel_vect (succ realdecls) params in
   let argsvect = extended_rel_vect 0 (CList.firstn (succ realdecls) ctx) in
   let pid = Id.of_string "P" in
-  let pdecl = make_assum (nameR pid) arity in
+  let pdecl = make_assum (make_annot (Name pid) ru) arity in
   let arity = lift 1 arity in
   let stepid = Id.of_string "step" in
   let recid = Id.of_string "rec" in
@@ -308,10 +312,10 @@ let derive_below env sigma ~poly (ind,univ as indu) =
       let ty = Vars.subst_instance_constr (EInstance.kind !evd univ) ty in
       let nargs = constructor_nrealargs env (ind, succ i) in
       let recarg = mkVar recid in
-      let args, _ = decompose_prod_assum !evd ty in
+      let args, _ = decompose_prod_decls !evd ty in
       let args, _ = List.chop (List.length args - params) args in
       let ty' = replace_term !evd (mkApp (mkIndU (ind,univ), rel_vect (-params) params)) recarg ty in
-      let args', _ = decompose_prod_assum !evd ty' in
+      let args', _ = decompose_prod_decls !evd ty' in
       let args', _ = List.chop (List.length args' - params) args' in
       let arg_tys = fst (List.fold_left (fun (acc, n) decl ->
         let t = get_type decl in
@@ -333,12 +337,12 @@ let derive_below env sigma ~poly (ind,univ as indu) =
       (* This wrapper checks if the argument is a recursive one,
        * and do the appropriate transformations if it is a product. *)
       let wrapper f = fun g (c, t) ->
-        let prem, res = decompose_prod_assum !evd t in
+        let prem, res = decompose_prod_decls !evd t in
         let t, args = decompose_app !evd res in
           if eq_constr !evd t recarg then
             let nprem = List.length prem in
             let elt = mkApp (lift nprem c, rel_vect 0 nprem) in
-            let args = Array.of_list (args @ [ elt ]) in
+            let args = Array.append args [| elt |] in
             let res, ty = f args nprem in
             let res = it_mkLambda_or_LetIn res prem in
             let ty = it_mkProd_or_LetIn ty prem in
@@ -375,7 +379,7 @@ let derive_below env sigma ~poly (ind,univ as indu) =
     in 
       lift 2 (it_mkLambda_or_LetIn caseB argbinders), lift 3 (it_mkLambda_or_LetIn caseb argbinders)
   in
-  let fixB = mkFix (([| realargs |], 0), ([| nameR recid |], [| arity |],
+  let fixB = mkFix (([| realargs |], 0), ([| make_annot (Name recid) ru |], [| arity |],
 				     [| subst_vars !evd [recid; pid] termB |])) in
   let bodyB = it_mkLambda_or_LetIn fixB (pdecl :: parambinders) in
   let id = add_prefix "Below_" (Nametab.basename_of_global (GlobRef.IndRef ind)) in
